=head1 NAME

ParCtl::Config - parctl configuration

=head1 SYNOPSIS

  read_configs($default_cfg_file, $user_cfg_file);

  adjust_offsets();

=head1 DESCRIPTION

This module contains parctl configuration and related functions.

=cut

package ParCtl::Config;

use strict;
use warnings;
use Carp;
use Config::Tiny;
use ParCtl::Time;
use Time::Piece;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(read_configs adjust_offsets);

=head1 VARIABLES

These package variables are populated by read_configs().  They expose
configurations for other modules.

=head2 $config

Reference to the configuration hash.  It has the same structure
(sections, keys and values) as the configuration INI files, converted
to Perl hash representation.  See Config::Tiny(3).

=head2 %flags

For convenience, some configuration values are transformed to boolean
flags:

  $ParCtl::Config::flags{parctl}   # true if parctl feature is on
  $ParCtl::Config::flags{datetime} # true if datetime feature is on
  $ParCtl::Config::flags{deadline} # true if slack mode is deadline
  $ParCtl::Config::flags{fixed}    # true if slack mode is fixed
  $ParCtl::Config::flags{utc}      # true if configured times are in UTC

These are easier to use than checking the corresponding configuration
values.

=head2 %times

For convenience, some configured times are converted from "HH:MM:SS"
format into seconds:

  $ParCtl::Config::times{slack}
  $ParCtl::Config::times{idleness}
  $ParCtl::Config::times{vulnerable}
  $ParCtl::Config::times{endangered}
  $ParCtl::Config::times{critical}
  $ParCtl::Config::times{reset}
  $ParCtl::Config::times{deadline}

=cut

our $config;
our %flags;
our %times;

=head1 FUNCTIONS

=head2 read_configs

  read_configs($default_config_file, $user_config_file);

Read the default configuration from $default_config_file, and user
configuration from $user_config_file, if it exists.  Configuration
files are expected to be in INI format; see Config::Tiny(3).  Croak if
$default_config_file does not exist, or in case of parsing errors.

The configuration can be accessed via $ParCtl::Config::config hashref
(see VARIABLES).  User configuration values override the corresponding
default values.

=cut

sub read_configs {
    my ($default_config_file, $user_config_file) = @_;
    $config = read_config($default_config_file);
    if (-f $user_config_file) {
        my $user_config = read_config($user_config_file);
        merge_configs($config, $user_config);
    }
    infer_flags($config);
    convert_times($config);
}

=head2 adjust_offsets

  adjust_offsets();

Convert the configured deadline and reset times to offsets in seconds
since UTC midnight.  See ParCtl::Time::time_offset() for details.  The
configured times are assumed to be in the local timezone, including
DST, unless $ParCtl::Config::flags{utc} is true.

Resulting offsets are stored into $ParCtl::Config::times{reset} and
$ParCtl::Config::times{deadline}.

=cut

sub adjust_offsets {
    $times{reset} = to_offset($config->{slack}{reset}, $flags{utc});
    $times{deadline} = to_offset($config->{slack}{deadline}, $flags{utc}) if ($flags{deadline});
}

=head1 NON-EXPORTED FUNCTIONS

=head2 read_config

  $text = read_config($filename);

Read and return configuration from $filename.

=cut

sub read_config {
    my $filename = shift;
    my $config = Config::Tiny->read($filename);
    $config or croak("cannot read configuration from $filename: " .
                     $Config::Tiny::errstr);
    return $config;
}

=head2 parse_config

  $config = parse_config($text);

Parse configuration $text from INI-format.  Return a hashref to the
configuration data.

This function is called only from tests.

=cut

sub parse_config {
    my $text = shift;
    return Config::Tiny->read_string($text);
}

=head2 merge_configs

  merge_configs($default_cfg, $user_cfg);

For each section/key/value pair defined in $user_cfg, add or replace
the corresponding section/key/value pair in $default_cfg.  In other
words, all user configurations take precedence over defaults.

=cut

sub merge_configs {
    my ($default, $user) = @_;
    for my $section (keys %$user) {
        my $defsect = $default->{$section} //= {};
        my $usersect = $user->{$section};
        @$defsect{keys %$usersect} = values %$usersect;
    }
}

=head2 infer_flags

  infer_flags($config);

Infer flags from $config.  See %flags in VARIABLES.

=cut

sub infer_flags {
    my $config = shift;
    $flags{parctl} = lc $config->{features}{parctl} eq "true";
    $flags{datetime} = lc $config->{features}{datetime} eq "true";
    $flags{fixed} = lc $config->{slack}{mode} eq "fixed";
    $flags{deadline} = lc $config->{slack}{mode} eq "deadline";
    $flags{utc} = lc $config->{slack}{utc} eq "true";
}

=head2 convert_times

  convert_times($config);

Convert slack and threshold times configured in "HH:MM:SS" format to
seconds.  See %times in VARIABLES.

=cut

sub convert_times {
    my $config = shift;
    $times{slack} = ParCtl::Time::str_to_secs($config->{slack}{initial});
    $times{idleness} = ParCtl::Time::str_to_secs($config->{limits}{idleness});
    $times{vulnerable} = ParCtl::Time::str_to_secs($config->{limits}{vulnerable});
    $times{endangered} = ParCtl::Time::str_to_secs($config->{limits}{endangered});
    $times{critical} = ParCtl::Time::str_to_secs($config->{limits}{critical});
}

=head2 to_offset

  to_offset($str, $utc);

Convert time string in "HH:MM:SS" format to an offset in seconds since
UTC midnight.  Adjust for the current timezone and DST unless $utc is
true.

=cut

sub to_offset {
    my ($str, $utc) = @_;
    my %hms = ParCtl::Time::str_to_hms($str);
    unless ($utc) {
        my $now = localtime;
        $hms{tzoffset} = $now->tzoffset;
    } else {
        $hms{tzoffset} = 0;
    }
    return ParCtl::Time::time_offset(%hms);
}

=head1 COPYRIGHT AND LICENSE

Copyright 2022 Petteri Hintsanen <petterih@iki.fi>

This module is free software; you can redistribute it and/or modify it
under the same terms as Perl itself.

=head1 SEE ALSO

parctl(1), ParCtl::Time(3), Config::Tiny(3).

=cut

1
