=head1 NAME

ParCtl::State - parctl state logic

=head1 SYNOPSIS

  add_callback(sub { say "state changed"; });

  toggle_consuming();

  toggle_grace();

  for (;;) { update_state(time()); say $state{slack}; sleep 1; }

=head1 DESCRIPTION

This module contains parctl state and related subroutines.  "State"
here means remaining slack, grace/consuming mode, and such things.

=cut

package ParCtl::State;

use v5.10;
use strict;
use warnings;
use ParCtl::Config;
use ParCtl::Time;
use X11::IdleTime;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw
    (add_callback
     reset_state
     update_state
     toggle_grace
     toggle_consuming);

use constant { IDLE => 0,
               LEAST_CONCERN => 1,
               VULNERABLE => 2,
               ENDANGERED => 3,
               CRITICAL => 4,
               EXTINCT => 5 };

=head1 VARIABLES

It's ok to read these package variables, but don't modify them
directly.  Use functions below for that.

=head2 %state

  %state =
    (current => undef,          # current parctl state
     slack => undef);           # remaining slack in seconds

Hash for the current state; updated by update_state().  State is one
of IDLE, LEAST_CONCERN, VULNERABLE, ENDANGERED, or CRITICAL.

=head2 %modes

  %modes =
    (consuming => 0,            # are we consuming?
     grace => 0);               # are we in grace mode?

Current modes.  These modify the normal state logic.  Modified by
toggle_consuming() and toggle_grace().

=cut

our %state =
    (current => undef,
     slack => undef);

our %modes =
    (consuming => 0,
     grace => 0);

my $reset_time;               # reset time offset since midnight
my @callbacks;                # coderefs called whenever state changes

=head1 FUNCTIONS

=head2 add_callback

  add_callback($coderef);

Add $coderef to the list of callbacks.  It will be called, without
arguments, after any state or mode change.

=cut

sub add_callback {
    my $sub = shift;
    push @callbacks, $sub;
}

=head2 reset_state

  reset_state();

Reset the state to undefined ("initial state").  This will cause
update_state to reset slack and time offsets.

=cut

sub reset_state {
    $state{current} = undef;
}

=head2 update_state

  update_state($now);

Maintain the current state w.r.t. current time $now: decrease slack,
reset if needed, transition into a new state, etc.  Fire callbacks
registered by add_callback() if %state changes.

This function should be called before trying to access %state for the
first time, and periodically after that.

=cut

sub update_state {
    my $now = shift;
    state $prev = $now;
    my $reset = (not defined $state{current}
                 or ($prev < $reset_time and $now >= $reset_time));
    if ($reset) {
        reset_slack($now);
    } elsif ($state{current} == EXTINCT) {
        # System was suspended but the reset time has not yet passed;
        # give some slack to prevent immediate re-suspension.
        $state{slack} = $ParCtl::Config::times{critical};
    } else {
        adjust_slack($prev, $now);
    }
    $prev = $now;
    transition();
}

=head2 toggle_grace

  toggle_grace();

Toggle grace mode.  In grace mode slack is not decreased.

=cut

sub toggle_grace {
    my $now = shift;
    $modes{grace} = !$modes{grace};
    notify();
}

=head2 toggle_consuming

  toggle_consuming();

Toggle consuming mode.  In consuming mode, parctl never goes to idle
state.

=cut

sub toggle_consuming {
    if ($ParCtl::Config::flags{fixed}) {
        $modes{consuming} = !$modes{consuming};
        notify();
    }
}

=head1 NON-EXPORTED FUNCTIONS

=head2 reset_slack

  reset_slack($now);

Reset the slack.  Exact behavior depends on the slack mode:

=over

=item * In slack mode, reset slack to the configured initial value.

=item * In deadline mode, calculate new slack from the current time to
the configured deadline.  If the new slack is less than the critical
limit, set it to the critical limit to prevent almost immediate
suspension.

=back

After adjusting slack update the current state accordingly.

=cut

sub reset_slack {
    my $now = shift;
    # Offsets are adjusted at each reset since daylight saving time or
    # even the time zone may change during program execution.
    ParCtl::Config::adjust_offsets();

    $reset_time = ParCtl::Time::time_from($now, $ParCtl::Config::times{reset});
    if ($ParCtl::Config::flags{deadline}) {
        $state{slack} = ParCtl::Time::time_to_deadline($now,
                                                       $ParCtl::Config::times{reset},
                                                       $ParCtl::Config::times{deadline});
        $state{slack} > $ParCtl::Config::times{critical}
            or $state{slack} = $ParCtl::Config::times{critical};
    } else {
        $state{slack} = $ParCtl::Config::times{slack};
    }
}

=head2 adjust_slack

  adjust_slack($prev, $now);

Decrease slack by difference between timestamps $prev and $now.  If
the difference is bigger than five seconds, assume that something out
of the ordinary (like temporary system suspend) has happened.  In that
case slack adjustment depends on the configured parctl mode:

=over

=item * In slack mode decrease slack by one second.

=item * In deadline mode reset slack.

=back

In idle or grace mode this function does nothing.

=cut

sub adjust_slack {
    my ($prev, $now) = @_;
    unless ($state{current} == IDLE or $modes{grace}) {
        my $diff = ($now - $prev);
        if ($diff > 5 and $ParCtl::Config::flags{fixed}) {
            decrease_slack(1);
        } elsif ($diff > 5 and $ParCtl::Config::flags{deadline}) {
            reset_slack($now);
        } else {
            decrease_slack($diff);
        }
    }
}

=head2 decrease_slack

  decrease_slack($amount);

Decrease slack by $amount seconds and update the current state.  Slack
goes never below zero.

=cut

sub decrease_slack {
    my $amount = shift;
    $state{slack} -= $amount;
    $state{slack} >= 0 or $state{slack} = 0;
}

=head2 transition

  transition();

Update the current state based on idleness and slack, and notify
watchers via callbacks.

If slack goes to zero, suspend the system.

=cut

sub transition {
    if (is_idle()) {
        $state{current} = IDLE;
    } elsif ($state{slack} <= 0) {
        $state{current} = EXTINCT;
    } elsif ($state{slack} <= $ParCtl::Config::times{critical}) {
        $state{current} = CRITICAL;
    } elsif ($state{slack} <= $ParCtl::Config::times{endangered}) {
        $state{current} = ENDANGERED;
    } elsif ($state{slack} <= $ParCtl::Config::times{vulnerable}) {
        $state{current} = VULNERABLE;
    } else {
        $state{current} = LEAST_CONCERN;
    }
    notify();
    suspend() if ($state{current} == EXTINCT);
}

=head2 notify

  notify();

Call all registered callbacks.

=cut

sub notify {
    foreach my $f (@callbacks) { $f->(); }
}

=head2 is_idle

  is_idle();

Return true if the system is currently idle, false otherwise.

The system is never considered to be idle if any of consuming, grace,
or deadline modes are active.

=cut

sub is_idle {
    return (GetIdleTime() >= $ParCtl::Config::times{idleness}
            and not $modes{consuming}
            and not $modes{grace}
            and not $ParCtl::Config::flags{deadline});
}

=head2 suspend

  suspend();

Suspend the system.

=cut

sub suspend {
    my $cmd = $ParCtl::Config::config->{suspend}{command};
    system($cmd) if $cmd;
}

=head1 COPYRIGHT AND LICENSE

Copyright 2022 Petteri Hintsanen <petterih@iki.fi>

This module is free software; you can redistribute it and/or modify it
under the same terms as Perl itself.

=head1 SEE ALSO

parctl(1)

=cut

1
