=head1 NAME

ParCtl::Time - parctl time related subroutines

=head1 SYNOPSIS

  $now = DateTime->new(year => 2022,
                      month => 1,
                      day => 29,
                      hour => 12,
                      time_zone => 'Europe/Helsinki');
  $lat = 60.167;
  $lon = 24.944;
  ($tod, $sunrise, $sunset) = day_or_night $now, $lat, $lon;

  str_to_hms("12:34:56"); # (hours => 12, minutes => 32, seconds => 56)
  str_to_secs("12:34:56"); # 45296

  time_from(86400, 3000, -1);     # 3000
  time_from(86400 + 3000, 4000);  # 90400

  time_offset(hours => 2, minutes => 30); # 9000
  time_offset(hours => 1, tzoffset => 7200); # 82800

  time_to_deadline(50, 30, 60);   # 10
  time_to_deadline(80, 30, 60);   # 0

=head1 DESCRIPTION

This module contains parctl subroutines that deal with time.

=cut

package ParCtl::Time;

use strict;
use warnings;
use Astro::Sunrise;
use Carp;
use DateTime;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(day_or_night
                    str_to_hms
                    str_to_secs
                    time_from
                    time_offset
                    time_to_deadline);

use constant { SECS_IN_DAY => 24 * 60 * 60,
               NIGHTTIME => 0,
               DAYTIME => 1,
               POLAR_NIGHT => 2,
               POLAR_DAY => 3 };

=head1 FUNCTIONS

=head2 day_or_night

  ($time, $sunrise, $sunset) = day_or_night $datetime, $lat, $lon

Find out whether it is day or night at DateTime object $datetime, and
when time of the next or previous sunrise and sunset.  $time is one of:

=over

=item * DAYTIME, in which case $sunrise is the time of the previous
sunrise and $sunset is the time of the next sunset.

=item * NIGHTTIME, in which case $sunrise is the time of the next
sunrise and $sunset is the time of the previous sunset.

=item * POLAR_DAY, when there is no nighttime and $sunrise and $sunset
are undefined.

=item * POLAR_NIGHT, when there is no daytime and $sunrise and $sunset
are undefined.

=back

=cut

sub day_or_night {
    my ($now, $lat, $lon) = @_; # pretend that $dt refers to the current time
    my ($sunrise, $sunset) = sunrise({day => $now->day,
                                      month => $now->month,
                                      year => $now->year,
                                      polar => "retval",
                                      tz => $now->offset / 3600,
                                      isdst => 0, # dt->offset includes DST
                                      lat => $lat,
                                      lon => $lon});

    # Is it polar day or night?
    if ($sunrise eq "day") { return (POLAR_DAY, undef, undef); }
    elsif ($sunrise eq "night") { return (POLAR_NIGHT, undef, undef); }

    my ($hour, $min) = ($sunrise =~ /(\d+):(\d+)/);
    my $sunrise_today = $now->clone->set(hour => $hour, minute => $min);

    # Is it before sunrise?
    if ($now < $sunrise_today) {
        # Calculate the previous sunset.
        my $yesterday = $now->clone->subtract(days => 1);
        ($sunrise, $sunset) = sunrise({day => $yesterday->day,
                                       month => $yesterday->month,
                                       year => $yesterday->year,
                                       tz => $yesterday->offset / 3600,
                                       isdst => 0,
                                       lat => $lat,
                                       lon => $lon});
        ($hour, $min) = ($sunset =~ /(\d+):(\d+)/);
        my $sunset_yesterday = $yesterday->clone->set(hour => $hour, minute => $min);
        return (NIGHTTIME, $sunrise_today, $sunset_yesterday);
    }

    # Is it before sunset?
    ($hour, $min) = ($sunset =~ /(\d+):(\d+)/);
    my $sunset_today = $now->clone->set(hour => $hour, minute => $min);
    if ($now < $sunset_today) {
        return (DAYTIME, $sunrise_today, $sunset_today);
    }

    # It is after sunset.  Calculate the next sunrise.
    my $tomorrow = $now->clone->add(days => 1);
    ($sunrise, $sunset) = sunrise({day => $tomorrow->day,
                                   month => $tomorrow->month,
                                   year => $tomorrow->year,
                                   tz => ($tomorrow->offset / 3600),
                                   isdst => 0,
                                   lat => $lat,
                                   lon => $lon});
    ($hour, $min) = ($sunrise =~ /(\d+):(\d+)/);
    my $sunrise_tomorrow = $tomorrow->clone->set(hour => $hour, minute => $min);
    return (NIGHTTIME, $sunrise_tomorrow, $sunset_today);
}

=head2 str_to_hms

  %hms = str_to_hms($str);

Parse hours, minutes and seconds from $str into %hms with keys
"hours", "minutes" and "seconds".  Only "HH:MM:SS" format is
recognized (leading zeroes are ok).  Values are not validated: they
can be negative or otherwise nonsensical.

Croak on parse error.

=cut

sub str_to_hms {
    my $str = shift;
    my ($hours, $mins, $secs) = ($str =~ /(\d+):(\d+):(\d+)/)
        or croak "bad time string '$str'";
    return (hours => int($hours),
            minutes => int($mins),
            seconds => int($secs));
}

=head2 str_to_secs

  $seconds = str_to_secs($str);

Parse hours, minutes and seconds from $str and convert to a total
number of seconds.  See str_to_hms() for limits on $str.

=cut

sub str_to_secs {
    my $str = shift;
    my %hms = str_to_hms($str);
    return 60 * 60 * $hms{hours} + 60 * $hms{minutes} + $hms{seconds};
}

=head2 time_from

  $time = time_from($now, $offset, $prev);

Return the next or previous absolute timestamp from $now given a
relative timestamp $offset.

Absolute timestamp means a Unix timestamp (seconds since Epoch).
Relative timestamp means a wall clock time as seconds after midnight
(see time_offset()).  It is normalized to range [0, 86400).

Example with 24-hour clock (all times in UTC):

  $now = 473428800;     # 1 Jan 1985, 12:00
  $offset = 10800;      # 03:00, regardless of date
  time_from($now, $offset, 1);  # 473396400 = 1 Jan 1985 03:00
  time_from($now, $offset);     # 473482800 = 2 Jan 1985 03:00

  $offset = 81000;      # 22:30, regardless of date
  time_from($now, $offset, 1);  # 473380200 = 31 Dec 1984 22:30
  time_from($now, $offset);     # 473466600 = 1 Dec 1985 22:30

  $offset = 43200;      # 12:00, regardless of date -- same time as $now
  time_from($now, $offset, 1);  # 473342400 = 31 Dec 1984 12:00
  time_from($now, $offset);     # 473515200 = 2 Jan 1985 12:00

=cut

sub time_from {
    my ($now, $offset, $prev) = @_;
    my $time;
    $offset = $offset % SECS_IN_DAY;
    if ($prev) {
        my $diff = ($now % SECS_IN_DAY) - $offset;
        if ($diff <= 0) { $diff = SECS_IN_DAY + $diff; }
        $time = $now - $diff;
    } else {
        my $diff = $offset - ($now % SECS_IN_DAY);
        if ($diff <= 0) { $diff = SECS_IN_DAY + $diff; }
        $time = $now + $diff;
    }
    return $time;
}

=head2 time_offset

  $offset = time_offset(
      hours => 12,
      minutes => 23,
      seconds => 45,
      tzoffset => 3600 # offset from UTC, in seconds
  )

Return the positive offset of the given time from UTC midnight, in
seconds.  If the time is not in UTC time, C<tzoffset> is subtracted
from it to get the corresponding UTC time.

All parameters have a default value of zero.

Examples:

  time_offset(hours => 1) # 3600
  time_offset(hours => 2, minutes => 30) # 9000
  time_offset(hours => 23) # 82800
  # which is the same as
  time_offset(hours => 1, tzoffset => 7200) # 82800

=cut

sub time_offset {
    my %args = (hours => 0,
                minutes => 0,
                seconds => 0,
                tzoffset => 0,
                @_);

    my $time = 60 * 60 * $args{hours} + 60 * $args{minutes} + $args{seconds};
    $time -= $args{tzoffset};
    $time %= SECS_IN_DAY;
    return $time;
}

=head2 time_to_deadline

  $diff = time_to_deadline($now, $reset_offset, $deadline_offset);

Return the number of seconds from $now to the $deadline, given that
reset it done at $reset.  All offsets are expected to be relative,
that is, wall clock times as seconds after midnight (see
time_offset()).  They are normalized to range [0, 86400).

Value of zero is returned in case $now is not between $reset and $deadline.

=cut

sub time_to_deadline {
    my ($now, $reset, $deadline) = @_;
    $reset != $deadline or croak "reset cannot be equal to deadline";
    $now %= 86400;
    $reset %= 86400;
    $deadline %= 86400;
    if ($now < $reset and $reset < $deadline) { return 0; }
    if ($reset <= $now and $now < $deadline) { return $deadline - $now; }
    if ($reset < $deadline and $deadline <= $now) { return 0; }
    if ($now < $deadline and $deadline < $reset) { return $deadline - $now; }
    if ($deadline <= $now and $now < $reset) { return 0; }
    if ($deadline < $reset and $reset <= $now) { return 86400 - $now + $deadline; }
    die "bug in time_to_deadline: now $now, reset $reset, deadline $deadline";
}

=head1 COPYRIGHT AND LICENSE

Copyright 2022 Petteri Hintsanen <petterih@iki.fi>

This module is free software; you can redistribute it and/or modify it
under the same terms as Perl itself.

=head1 SEE ALSO

parctl(1)

=cut

1
