=head1 NAME

ParCtl::Tk - parctl Tk subroutines

=head1 SYNOPSIS

  main();

=head1 DESCRIPTION

This module implements Perl/Tk GUI and the main event loop for parctl.

=cut

package ParCtl::Tk;

use v5.10;
use strict;
use warnings;
use DateTime;
use ParCtl::Config;
use ParCtl::State;
use ParCtl::Time;
use POSIX qw(strftime);
use Time::Piece;
use Tk::Balloon;
use Tk::MsgBox;
use Tk;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(main);

# References to Tk objects.
my %widgets;

# Balloon (tooltip) strings.
my %tooltips = (date => undef,
                time => undef,
                slack => undef);

my $timezone = DateTime::TimeZone->new(name => 'local');

=head1 FUNCTIONS

=head2 main

Initialize signals, widgets and parctl state, and start the Tk event
loop.

=cut

sub main {
    $SIG{INT} = \&Tk::exit;
    $SIG{TERM} = \&Tk::exit;
    make_widgets();
    ParCtl::State::add_callback(\&state_changed);
    $widgets{window}->repeat(1000, \&update);
    update();
    MainLoop();
}

=head1 NON-EXPORTED FUNCTIONS

=head2 make_widgets

  make_widgets();

Create all Tk widgets and store them into %widgets.  Also bind mouse
buttons to mode toggle functions.

=cut

sub make_widgets {
    make_windows();
    make_fonts();

    if ($ParCtl::Config::flags{datetime}) {
        make_datetime_widgets();
    }

    if ($ParCtl::Config::flags{parctl}) {
        make_slack_widgets();
        $widgets{window}->bind('<Button-2>', sub { ParCtl::State::toggle_grace() });
        $widgets{window}->bind('<Button-3>', sub { ParCtl::State::toggle_consuming() });
    }

    set_background($ParCtl::Config::config->{colors}{background});
}

=head2 make_windows

  make_windows();

Create main window and slack warning dialog.

=cut

sub make_windows {
    $widgets{window} = MainWindow->new(-title => "Parental Control");
    $ParCtl::Config::config->{colors}{background} //= $widgets{window}->cget("-background");
    $ParCtl::Config::config->{colors}{foreground} //= $widgets{window}->cget("-foreground");
    $widgets{warning} = $widgets{window}->MsgBox(-icon => 'warning',
                                                 -title => "Warning",
                                                 -message => "You are running out of slack!",
                                                 -type => 'ok');
}

=head2 make_fonts

  make_fonts();

Create Tk font objects for labels.

=cut

sub make_fonts {
    my %fontargs;
    my %balloon_fontargs;
    my $fonts = $ParCtl::Config::config->{fonts};
    $fontargs{-family} = $fonts->{font_family} if $fonts->{font_family};
    $fontargs{-size} = $fonts->{font_size} if $fonts->{font_size};
    $balloon_fontargs{-family} = $fonts->{tooltip_font_family} if $fonts->{tooltip_font_family};
    $balloon_fontargs{-size} = $fonts->{tooltip_font_size} if $fonts->{tooltip_font_size};
    $widgets{font} = $widgets{window}->fontCreate('C_font', %fontargs);
    $widgets{consume_font} = $widgets{window}->fontCreate('C_consumefont',
                                                          -weight => 'bold',
                                                          %fontargs);
    $widgets{balloon_font} = $widgets{window}->fontCreate('C_balloonfont',
                                                          %balloon_fontargs);
}

=head2 make_datetime_widges

  make_datetime_widgets();

Create date and time labels and tooltips.

=cut

sub make_datetime_widgets {
    my %labelargs = label_style();
    my %balloonargs = balloon_style();

    $widgets{date_label} = $widgets{window}->Label(%labelargs);
    $widgets{time_label} = $widgets{window}->Label(%labelargs);
    $widgets{date_label}->pack(-expand => 1, -anchor => "s");
    $widgets{time_label}->pack(-expand => 1, -anchor => "n");

    $balloonargs{-postcommand} = \&update_date_tooltip;
    $widgets{date_balloon} = $widgets{window}->Balloon(%balloonargs);
    $widgets{date_balloon}->attach($widgets{date_label},
                                   -balloonmsg => \$tooltips{date});

    my $location = $ParCtl::Config::config->{location};
    if ($location->{latitude} && $location->{longitude}) {
        $balloonargs{-postcommand} = \&update_time_tooltip;
        $widgets{time_balloon} = $widgets{window}->Balloon(%balloonargs);
        $widgets{time_balloon}->attach($widgets{time_label},
                                       -balloonmsg => \$tooltips{time});
    }
}

=head2 make_slack_widgets

  make_slack_widgets();

Create slack label and tooltip.

=cut

sub make_slack_widgets {
    my %labelargs = label_style();
    my %balloonargs = balloon_style();

    $widgets{slack_label} = $widgets{window}->Label(%labelargs);
    $widgets{slack_label}->pack(-expand => 1);
    $balloonargs{-postcommand} = \&update_slack_tooltip;
    $widgets{slack_balloon} = $widgets{window}->Balloon(%balloonargs);
    $widgets{slack_balloon}->attach($widgets{slack_label},
                                    -balloonmsg => \$tooltips{slack});
}

=head2 label_style

  %style = label_style();

Create and return named argument hash for creating labels.

=cut

sub label_style {
    my $colors = $ParCtl::Config::config->{colors};
    my %labelargs = (-background => $colors->{background},
                     -font => $widgets{font});
    $labelargs{-foreground} = $colors->{foreground};
    return %labelargs;
}

=head2 balloon_style

  %style = balloon_style();

Create and return named argument hash for creating tooltips (balloons
in Tk jargon).

=cut

sub balloon_style {
    my %balloonargs;
    my $colors = $ParCtl::Config::config->{colors};
    $balloonargs{-background} = $colors->{tooltip_background} if $colors->{tooltip_background};
    $balloonargs{-foreground} = $colors->{tooltip_foreground} if $colors->{tooltip_foreground};
    $balloonargs{-font} = $widgets{balloon_font};
    return %balloonargs;
}

=head2 update

  update();

Update date, time and state.  This function is called once per second.

=cut

sub update {
    if ($ParCtl::Config::flags{datetime}) { update_datetime(); }
    if ($ParCtl::Config::flags{parctl}) { ParCtl::State::update_state(time()); }
}

=head2 state_changed

  state_changed();

Respond to state change by updating widgets.  Warn the user if state
becomes critical.

=cut

sub state_changed {
    update_background();
    update_slack();
    warn_if_critical();
}

=head2 update_background

  update_background();

Set the background color according to the current state.

=cut

sub update_background {
    my $state = $ParCtl::State::state{current};
    my $colors = $ParCtl::Config::config->{colors};
    state $alarmcol = $colors->{endangered};

    if ($ParCtl::State::modes{grace}) {
        set_background($colors->{grace});
    } elsif ($state == ParCtl::State::IDLE) {
        set_background($colors->{idle});
    } elsif ($state == ParCtl::State::VULNERABLE) {
        set_background($colors->{vulnerable});
    } elsif ($state == ParCtl::State::ENDANGERED) {
        set_background($colors->{endangered});
    } elsif ($state == ParCtl::State::CRITICAL) {
        $alarmcol = $alarmcol eq $colors->{endangered} ?
            $colors->{vulnerable} : $colors->{endangered};
        set_background($alarmcol);
    } else {
        set_background($colors->{background});
    }
}

=head2 set_background

  set_background($color);

Set background color for all widgets to $color.

=cut

sub set_background {
    my $bg = shift;
    $widgets{window}->configure(-background => $bg);
    if ($widgets{slack_label}) {
        $widgets{slack_label}->configure(-background => $bg);
    }
    if ($widgets{date_label}) {
        $widgets{date_label}->configure(-background => $bg);
    }
    if ($widgets{time_label}) {
        $widgets{time_label}->configure(-background => $bg);
    }
}

=head2 update_slack

  update_slack();

Update slack widget.

=cut

sub update_slack {
    if ($ParCtl::State::modes{consuming}) {
        $widgets{slack_label}->configure(-font => $widgets{consume_font});
    } else {
        $widgets{slack_label}->configure(-font => $widgets{font});
    }
    $widgets{slack_label}->configure(-text => format_slack($ParCtl::State::state{slack}));
}

=head2 warn_if_critical

  warn_if_critical();

Display warning dialog if critical state was entered for the first
time.

=cut

sub warn_if_critical {
    state $warned = 0;
    if ($ParCtl::State::state{current} == ParCtl::State::CRITICAL
        and not $ParCtl::State::modes{grace}) {
        unless ($warned) {
            $warned = 1;
            $widgets{window}->after(0, sub { $widgets{warning}->Show(); });
        }
    } else {
        $warned = 0;
    }
}

=head2 update_date_tooltip

  update_date_tooltip();

Update date tooltip with the current date and week.  Display age in
days if date of birth is configured.

=cut

sub update_date_tooltip {
    my $now = DateTime->now;
    $now->set_time_zone($timezone);
    my $tt = sprintf("Today is %s, %s %s %s, week %s.",
                     $now->day_name,
                     $now->day,
                     $now->month_name,
                     $now->year,
                     $now->week_number);

    my $dobcfg = $ParCtl::Config::config->{date_of_birth};
    if ($dobcfg && $dobcfg->{day} && $dobcfg->{month} && $dobcfg->{year}) {
        my $dob = DateTime->new(day => $dobcfg->{day},
                                month => $dobcfg->{month},
                                year => $dobcfg->{year});
        my $days = $dob->delta_days($now)->{days} + 1;
        $tt .= sprintf("\nIt is %s day of your life.  Spend it wisely.",
                       to_ordinal($days));
    }
    $tooltips{date} = $tt;
}

=head2 update_time_tooltip

  update_time_tooltip();

Update time tooltip with time of day and time of the next sunset or
sunrise.

=cut

sub update_time_tooltip {
    my $location = $ParCtl::Config::config->{location};
    my $now = DateTime->now(time_zone => $timezone);
    my ($lat, $lon) = ($location->{latitude}, $location->{longitude});
    my ($tod, $sunrise, $sunset) = ParCtl::Time::day_or_night($now, $lat, $lon);
    if ($tod eq ParCtl::Time::NIGHTTIME) {
        my $tod_str = sprintf("It is %s", tod_to_str($tod));
        my ($sunset_str, $from_sunset) = make_duration_str($now, $sunset, "from sunset");
        my ($sunrise_str, $to_sunrise) = make_duration_str($now, $sunrise, "to sunrise");
        my $darkness = make_total_duration_str($from_sunset, $to_sunrise, "darkness");
        $tooltips{time} = sprintf("%s. %s\n%s\n%s", $tod_str, $darkness, $sunset_str, $sunrise_str);
    } elsif ($tod eq ParCtl::Time::DAYTIME) {
        my $tod_str = sprintf("It is %s", tod_to_str($tod));
        my ($sunrise_str, $from_sunrise) = make_duration_str($now, $sunrise, "from sunrise");
        my ($sunset_str, $to_sunset) = make_duration_str($now, $sunset, "to sunset");
        my $daylight = make_total_duration_str($from_sunrise, $to_sunset, "daylight");
        $tooltips{time} = sprintf("%s. %s\n%s\n%s", $tod_str, $daylight, $sunrise_str, $sunset_str);
    } else {
        # Polar day or night.
        $tooltips{time} = sprintf("It is %s.", tod_to_str($tod));
    }
}

=head2 tod_to_str

  $str = tod_to_str($time_of_day);

Convert time of day constant to string and return it.

=cut

sub tod_to_str {
    my $tod = shift;
    if ($tod == ParCtl::Time::NIGHTTIME) { return "nighttime"; }
    elsif ($tod == ParCtl::Time::DAYTIME) { return "daytime"; }
    elsif ($tod == ParCtl::Time::POLAR_NIGHT) { return "polar night"; }
    elsif ($tod == ParCtl::Time::POLAR_DAY) { return "polar day"; }
    else { die "bad time of day $tod"; }
}

=head2 make_duration_str

  ($str, $duration) = make_duration_str($now, $then, $event);

Calculate the duration between DateTime objects $now and $then, and
format it into a string with $event.

Return the formatted string and the duration as DateTime::Duration
object.

See update_time_tooltip().

=cut

sub make_duration_str {
    my ($now, $then, $event) = @_;

    my $to_then = $now->delta_ms($then);
    my ($hours, $mins) = (int($to_then->{minutes} / 60),
                          $to_then->{minutes} % 60);

    my $str;
    $str = "$hours h " if ($hours);
    $str .= "$mins min " if ($mins);
    $str .= "$event";
    my $time = strftime($ParCtl::Config::config->{formats}{time},
                        $then->sec,
                        $then->min,
                        $then->hour,
                        $then->mday,
                        $then->mon - 1,
                        $then->year - 1900);
    $str .= " ($time)";
    return ($str, $to_then);
}

=head2 make_total_duration_str

  $str = make_total_duration_str($from, $to, $event);

Calculate total duration from DateTime::Duration objects $from and $to
and format that into a string with $event.  Return the formatted
string.

See update_time_tooltip().

=cut

sub make_total_duration_str {
    my ($from, $to, $event) = @_;
    my $total_mins = $from->{minutes} + $to->{minutes};
    my ($hours, $mins) = (int($total_mins / 60), $total_mins % 60);
    my $str;
    $str = "$hours h " if ($hours);
    $str .= "$mins min " if ($mins);
    $str .= "$event";
    return $str;
}

=head2 update_slack_tooltip

  update_slack_tooltip();

Update slack tooltip.  Display the time of shutdown in configured time
format.

=cut

sub update_slack_tooltip {
    my $now = localtime(time() + $ParCtl::State::state{slack});
    my $time = strftime($ParCtl::Config::config->{formats}{time},
                        $now->sec,
                        $now->min,
                        $now->hour,
                        $now->mday,
                        $now->mon - 1,
                        $now->year - 1900);
    unless ($ParCtl::State::modes{grace}) {
        my $tt = "Suspend at $time";
        $tooltips{slack} = $tt;
    } else {
        $tooltips{slack} = "Grace mode on";
    }
}

=head2 update_datetime

  update_datetime();

Update date and time labels with the current date and time.  Use
configured date and time formats.

=cut

sub update_datetime {
    my $now = localtime;
    my $date = strftime($ParCtl::Config::config->{formats}{date},
                        $now->sec,
                        $now->min,
                        $now->hour,
                        $now->mday,
                        $now->mon - 1,
                        $now->year - 1900);
    my $time = strftime($ParCtl::Config::config->{formats}{time},
                        $now->sec,
                        $now->min,
                        $now->hour,
                        $now->mday,
                        $now->mon - 1,
                        $now->year - 1900);
    $widgets{date_label}->configure(-text => $date);
    $widgets{time_label}->configure(-text => $time);
}

=head2 format_slack

  $str = format_slack($seconds);

Convert slack given in $seconds to "HH:MM" (if more than one hour
slack remaing) or "MM:SS" format (if less than one hour slack
remaining).

Both formats are configurable.

=cut

sub format_slack {
    my $seconds = shift;
    $seconds = int($seconds + 0.5);
    my $hours = int($seconds / 60 / 60);
    my $mins = int($seconds / 60 % 60);
    my $secs = $seconds % 60;
    if ($hours > 0) {
        return strftime($ParCtl::Config::config->{formats}{slack_hm},
                        $secs, $mins, $hours, 0, 0, 0);
    } else {
        return strftime($ParCtl::Config::config->{formats}{slack_ms},
                        $secs, $mins, $hours, 0, 0, 0);
    }
}

=head2 to_ordinal

  $str = to_ordinal($int);

Convert $int to an ordinal number string and return that.

=cut

sub to_ordinal {
    my $num = shift;
    my $tens = $num % 100;
    if ($tens >= 10 && $tens <= 19) {
        return "${num}th";
    } else {
        my $units = $tens % 10;
        if ($units == 1) { return "${num}st"; }
        elsif ($units == 2) { return "${num}nd"; }
        elsif ($units == 3) { return "${num}rd"; }
        else { return "${num}th"; };
    }
}

=head1 COPYRIGHT AND LICENSE

Copyright 2022 Petteri Hintsanen <petterih@iki.fi>

This module is free software; you can redistribute it and/or modify it
under the same terms as Perl itself.

=head1 SEE ALSO

parctl(1)

=cut

1
