#!/usr/bin/env perl

use warnings;
use strict;
use FindBin;
use Test::More;

use lib "$FindBin::Bin/../lib";
use ParCtl::State qw(update_state toggle_consuming toggle_grace);
use ParCtl::Config;

# Test config template.  This will be mutilated in deadline tests
# below.
my $test_cfg = <<'EOF';
[features]
parctl=true
datetime=false

[slack]
mode=fixed
initial=00:00:10
reset=00:00:08
utc=true

[limits]
idleness=00:01:00
vulnerable=00:00:08
endangered=00:00:06
critical=00:00:04

[suspend]
suspend_cmd=
EOF

# Fake read_configs()
$ParCtl::Config::config = ParCtl::Config::parse_config($test_cfg);
ParCtl::Config::infer_flags($ParCtl::Config::config);
ParCtl::Config::convert_times($ParCtl::Config::config);

#
# Fixed slack mode
#

# Uninitialized state, will cause reset.
$ParCtl::State::state{current} = undef;
update_state(0);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
is($ParCtl::State::state{slack}, 10);
update_state(1);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
is($ParCtl::State::state{slack}, 9);
update_state(2);
is($ParCtl::State::state{current}, ParCtl::State::VULNERABLE);
is($ParCtl::State::state{slack}, 8);
update_state(3);
is($ParCtl::State::state{current}, ParCtl::State::VULNERABLE);
is($ParCtl::State::state{slack}, 7);
update_state(4);
is($ParCtl::State::state{current}, ParCtl::State::ENDANGERED);
is($ParCtl::State::state{slack}, 6);
update_state(5);
is($ParCtl::State::state{current}, ParCtl::State::ENDANGERED);
is($ParCtl::State::state{slack}, 5);
update_state(6);
is($ParCtl::State::state{current}, ParCtl::State::CRITICAL);
is($ParCtl::State::state{slack}, 4);
update_state(7);
is($ParCtl::State::state{current}, ParCtl::State::CRITICAL);
is($ParCtl::State::state{slack}, 3);

# Reset time.
update_state(8);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
is($ParCtl::State::state{slack}, 10);
update_state(9);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
is($ParCtl::State::state{slack}, 9);

# Jump more than five seconds, but to a point before next reset, which
# is at 86408.
update_state(1000);
is($ParCtl::State::state{current}, ParCtl::State::VULNERABLE);
is($ParCtl::State::state{slack}, 8);

# Do longer than one second jumps to save some lines.
update_state(1002);
is($ParCtl::State::state{current}, ParCtl::State::ENDANGERED);
is($ParCtl::State::state{slack}, 6);
update_state(1004);
is($ParCtl::State::state{current}, ParCtl::State::CRITICAL);
is($ParCtl::State::state{slack}, 4);
update_state(1007);
is($ParCtl::State::state{current}, ParCtl::State::CRITICAL);
is($ParCtl::State::state{slack}, 1);

# Slack exhausted, time to suspend.
update_state(1008);
is($ParCtl::State::state{current}, ParCtl::State::EXTINCT);
is($ParCtl::State::state{slack}, 0);

# Updating before reset time will result in critical slack.
update_state(1020);
is($ParCtl::State::state{current}, ParCtl::State::CRITICAL);
is($ParCtl::State::state{slack}, 4);

# Almost there in the reset time.
update_state(86407);
is($ParCtl::State::state{current}, ParCtl::State::CRITICAL);
is($ParCtl::State::state{slack}, 3);

# Now reset.
update_state(86408);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
is($ParCtl::State::state{slack}, 10);

# Grace test.
update_state(86409);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
is($ParCtl::State::state{slack}, 9);

toggle_grace();
update_state(86411);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
is($ParCtl::State::state{slack}, 9);

update_state(86412);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
is($ParCtl::State::state{slack}, 9);

toggle_grace();
update_state(86413);
is($ParCtl::State::state{current}, ParCtl::State::VULNERABLE);
is($ParCtl::State::state{slack}, 8);

# Idleness and consume mode tests.
$ParCtl::Config::times{idleness} = 0;
update_state(86414);
# Note that slack is always adjusted before state transitions.
is($ParCtl::State::state{current}, ParCtl::State::IDLE);
is($ParCtl::State::state{slack}, 7);
update_state(86416);
is($ParCtl::State::state{current}, ParCtl::State::IDLE);
is($ParCtl::State::state{slack}, 7);
toggle_consuming();
update_state(86417);
is($ParCtl::State::state{current}, ParCtl::State::VULNERABLE);
is($ParCtl::State::state{slack}, 7);
update_state(86418);
is($ParCtl::State::state{current}, ParCtl::State::ENDANGERED);
is($ParCtl::State::state{slack}, 6);
toggle_consuming();
$ParCtl::Config::times{idleness} = 60;

#
# Deadline mode
#

$ParCtl::Config::flags{fixed} = 0;
$ParCtl::Config::flags{deadline} = 1;
$ParCtl::Config::config->{slack}{reset} = "00:10:00";
$ParCtl::Config::config->{slack}{deadline} = "00:01:00";
$ParCtl::Config::config->{limits}{vulnerable} = "00:00:45";
$ParCtl::Config::config->{limits}{endangered} = "00:00:30";
$ParCtl::Config::config->{limits}{critical} = "00:00:15";
ParCtl::Config::convert_times($ParCtl::Config::config);

# Uninitialized state.
$ParCtl::State::state{current} = undef;
update_state(0);
is($ParCtl::State::state{slack}, 60);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
update_state(1);
is($ParCtl::State::state{slack}, 59);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
# Jump over five seconds, which will result in slack reset.
# In this case there will be ample amount of slack remaining.
update_state(10);
is($ParCtl::State::state{slack}, 50);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
# Jump again over five seconds, but this time there will be less than
# critical slack.  In this case slack will be reset to the critical
# limit.
update_state(48);
is($ParCtl::State::state{slack}, 15);
is($ParCtl::State::state{current}, ParCtl::State::CRITICAL);
# Jump over the deadline, which is at 63, but to a point before next
# reset, which is at 600.  This will reset to a critical slack to
# prevent immediate shutdown.
update_state(100);
is($ParCtl::State::state{slack}, 15);
is($ParCtl::State::state{current}, ParCtl::State::CRITICAL);
update_state(102);
is($ParCtl::State::state{slack}, 13);
is($ParCtl::State::state{current}, ParCtl::State::CRITICAL);
update_state(105);
is($ParCtl::State::state{slack}, 10);
update_state(109);
is($ParCtl::State::state{slack}, 6);
update_state(113);
is($ParCtl::State::state{slack}, 2);
update_state(114);
is($ParCtl::State::state{slack}, 1);
update_state(115);
is($ParCtl::State::state{slack}, 0);
is($ParCtl::State::state{current}, ParCtl::State::EXTINCT);
update_state(116);
is($ParCtl::State::state{slack}, 15);
is($ParCtl::State::state{current}, ParCtl::State::CRITICAL);
update_state(216);
is($ParCtl::State::state{slack}, 15);
is($ParCtl::State::state{current}, ParCtl::State::CRITICAL);

# Not quite reset yet...
update_state(599);
is($ParCtl::State::state{slack}, 15);
is($ParCtl::State::state{current}, ParCtl::State::CRITICAL);
# Now at reset.  Next deadline is at 86460.
update_state(600);
is($ParCtl::State::state{slack}, 86460 - 600);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);

# Grace test.
update_state(601);
is($ParCtl::State::state{slack}, 85859);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
toggle_grace();
update_state(603);
is($ParCtl::State::state{slack}, 85859);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
update_state(605);
is($ParCtl::State::state{slack}, 85859);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);
toggle_grace();
update_state(607);
is($ParCtl::State::state{slack}, 85857);
is($ParCtl::State::state{current}, ParCtl::State::LEAST_CONCERN);

done_testing();
