#!/usr/bin/env perl

use warnings;
use strict;
use FindBin;
use Test::More;

use lib "$FindBin::Bin/../lib";
use ParCtl::Time qw(time_offset
                    time_from
                    time_to_deadline
                    day_or_night
                    str_to_hms
                    str_to_secs);

use constant { MINUTE => 60,
               HOUR => 60 * 60,
               DAY => 24 * 60 * 60 };

#
# day_or_night
#

# These are known good values.
my $sunrise = DateTime->new(year => 2022,
                            month => 1,
                            day => 29,
                            hour => 8,
                            minute => 42,
                            second => 0,
                            time_zone => 'Europe/Helsinki');

my $sunset = DateTime->new(year => 2022,
                           month => 1,
                           day => 29,
                           hour => 16,
                           minute => 25,
                           second => 0,
                           time_zone => 'Europe/Helsinki');

my $prev_sunset = DateTime->new(year => 2022,
                                month => 1,
                                day => 28,
                                hour => 16,
                                minute => 22,
                                second => 0,
                                time_zone => 'Europe/Helsinki');

my $next_sunrise = DateTime->new(year => 2022,
                                 month => 1,
                                 day => 30,
                                 hour => 8,
                                 minute => 39,
                                 second => 0,
                                 time_zone => 'Europe/Helsinki');

# This "now" will float around in the tests below.
my $now = DateTime->new(year => 2022,
                        month => 1,
                        day => 29,
                        time_zone => 'Europe/Helsinki');
my $lat = 60.166640739;       # Helsinki
my $lon = 24.943536799;

# Well before sunrise.
$now->set(hour => 5, minute => 30);
my ($tod, $sr, $ss) = day_or_night $now, $lat, $lon;
is($tod, ParCtl::Time::NIGHTTIME);
is($sr, $sunrise);
is($ss, $prev_sunset);

# Right before sunrise.
$now->set(hour => 8, minute => 41);
($tod, $sr, $ss) = day_or_night $now, $lat, $lon;
is($tod, ParCtl::Time::NIGHTTIME);
is($sr, $sunrise);
is($ss, $prev_sunset);

# At sunrise.
$now->set(hour => 8, minute => 42);
($tod, $sr, $ss) = day_or_night $now, $lat, $lon;
is($tod, ParCtl::Time::DAYTIME);
is($sr, $sunrise);
is($ss, $sunset);

# At noon.
$now->set(hour => 12, minute => 0);
($tod, $sr, $ss) = day_or_night $now, $lat, $lon;
is($tod, ParCtl::Time::DAYTIME);
is($sr, $sunrise);
is($ss, $sunset);

# Right before sunset.
$now->set(hour => 16, minute => 24);
($tod, $sr, $ss) = day_or_night $now, $lat, $lon;
is($tod, ParCtl::Time::DAYTIME);
is($sr, $sunrise);
is($ss, $sunset);

# At sunset.
$now->set(hour => 16, minute => 25);
($tod, $sr, $ss) = day_or_night $now, $lat, $lon;
is($tod, ParCtl::Time::NIGHTTIME);
is($sr, $next_sunrise);
is($ss, $sunset);

# At evening.
$now->set(hour => 18, minute => 0);
($tod, $sr, $ss) = day_or_night $now, $lat, $lon;
is($tod, ParCtl::Time::NIGHTTIME);
is($sr, $next_sunrise);
is($ss, $sunset);

# Test for polar night.
$lat = 78.2166658;            # Longyearbyen
$lon = 15.5499978;
($tod, $ss, $sr) = day_or_night $now, $lat, $lon;
is($tod, ParCtl::Time::POLAR_NIGHT);
is($ss, undef);
is($sr, undef);

# Test for polar day.
$lat = -78.466139;            # Vostok station
$lon = 106.84825;
($tod, $ss, $sr) = day_or_night $now, $lat, $lon;
is($tod, ParCtl::Time::POLAR_DAY);
is($ss, undef);
is($sr, undef);

# DST tests.  DST became effective at 31 Mar 2022 03:00 EEST
$now = DateTime->new(year => 2022,
                     month => 3,
                     day => 31,
                     hour => 2,
                     minute => 0,
                     time_zone => 'Europe/Helsinki');
$lat = 60.166640739;          # Helsinki
$lon = 24.943536799;

($tod, $sr, $ss) = day_or_night $now, $lat, $lon;
is($tod, ParCtl::Time::NIGHTTIME);
# Set $now to the next sunrise (06:48 in EEST).
$now->set(hour => 6, minute => 48);
is($sr, $now);

($tod, $sr, $ss) = day_or_night $now, $lat, $lon;
is($tod, ParCtl::Time::DAYTIME);
# Set $now to the next sunset (20:01 is EEST).
$now->set(hour => 20, minute => 01);
is($ss, $now);

($tod, $sr, $ss) = day_or_night $now, $lat, $lon;
is($tod, ParCtl::Time::NIGHTTIME);
# Set $now to the next sunrise (1 Apr 2022 06:45 EEST).
$now->set(day => 1, month => 4, hour => 6, minute => 45);
is($sr, $now);

#
# str_to_hms
#

my %hms = str_to_hms("00:00:00");
my %exp = ("hours" => 0, "minutes" => 0, "seconds" => 0);
is_deeply(\%hms, \%exp);

%hms = str_to_hms("12:34:56");
%exp = ("hours" => 12, "minutes" => 34, "seconds" => 56);
is_deeply(\%hms, \%exp);

#
# str_to_secs
#

is(str_to_secs("12:34:56"), 45296);
is(str_to_secs("02:00:00"), 7200);
is(str_to_secs("00:15:00"), 900);
is(str_to_secs("00:10:00"), 600);
is(str_to_secs("00:02:00"), 120);
is(str_to_secs("00:00:30"), 30);

#
# time_from
#

is(time_from(86400, 3000, -1), 3000);
is(time_from(86400 + 3000, 4000), 90400);
is(time_from(9, 8), 86400 + 8);

$now = 5 * DAY + 2 * HOUR;
my $offset = time_offset(hours => 18);
is(time_from($now, $offset, 1), $now - 8 * HOUR);
is(time_from($now, $offset), $now + 16 * HOUR);

$now = MINUTE;
$offset = 23 * HOUR;
is(time_from($now, $offset, 1), $now - HOUR - MINUTE);
is(time_from($now, $offset - MINUTE, 1), $now - HOUR - 2 * MINUTE);
is(time_from($now, $offset), $now + 23 * HOUR - 1 * MINUTE);
is(time_from($now, $offset + MINUTE), $now + 23 * HOUR);

$now = -DAY;
$offset = 0;
is(time_from($now, $offset), 0);
is(time_from($now, $offset, 1), -2 * DAY);
$offset = 2 * DAY;
is(time_from($now, $offset), 0);
$offset = 6 * HOUR;
is(time_from($now, $offset, 1), $now - 18 * HOUR);
is(time_from($now, $offset), $now + 6 * HOUR);

$now = 7 * DAY + 18 * HOUR;
$offset = time_offset(hours => 3, minutes => 30);
is(time_from($now, $offset), $now + 9 * HOUR + 30 * MINUTE);
is(time_from($now, $offset, 1), $now - 14 * HOUR - 30 * MINUTE);

$now = 473428800;               # 1 Jan 1985, 12:00
$offset = 10800;                # 03:00, regardless of date
is(time_from($now, $offset, 1), 473396400); # 1 Jan 1985 03:00
is(time_from($now, $offset), 473482800);    # 2 Jan 1985 03:00

$offset = 81000;                # 22:30, regardless of date
is(time_from($now, $offset, 1), 473380200); # 31 Dec 1984 22:30
is(time_from($now, $offset), 473466600);    # 1 Dec 1985 22:30

$offset = 43200;      # 12:00, regardless of date -- same time as $now
is(time_from($now, $offset, 1), 473342400); # 31 Dec 1984 12:00
is(time_from($now, $offset), 473515200);    # 2 Jan 1985 12:00

#
# time_offset
#

is(time_offset(), 0);
is(time_offset(hours => 1), 3600);
is(time_offset(hours => 2, minutes => 30), 9000);
is(time_offset(hours => 2, minutes => 30, seconds => 45), 9045);
is(time_offset(hours => 22, seconds => 1, tzoffset => -7200), 1);
is(time_offset(hours => 23), 82800);
is(time_offset(hours => 24), 0);
is(time_offset(hours => 1, tzoffset => 7200), 82800);
is(time_offset(hours => 21, tzoffset => -7200), 82800);
is(time_offset(hours => -1), 82800);
is(time_offset(hours => -1, tzoffset => -3600), 0);
is(time_offset(hours => -1, tzoffset => -3600), 0);

#
# time_to_deadline
#

$now = 6 * DAY + HOUR;
my $reset = time_offset(hours => 21);
my $deadline = time_offset(hours => 3);
is(time_to_deadline($now, $reset, $deadline), 2 * HOUR);
$now = 6 * DAY + 6 * HOUR;
is(time_to_deadline($now, $reset, $deadline), 0);
is(time_to_deadline(80, 30, 60), 0);
done_testing();
